﻿namespace DataModel
{
  using System.Linq;
  using System.Security;
  using System.Transactions;
  using Microsoft.VisualStudio.TestTools.UnitTesting;

  [TestClass]
  public class NorthwindEntitiesTest
  {
    [TestMethod]
    [ExpectedException(typeof(SecurityException))]
    public void SuppliersThrowsSecurityExceptionWhenUserIsNotAllowedToRead()
    {
      using (new TransactionScope())
      using (var context = new NorthwindEntities())
      using (new GenericPrincipalScope("Celene", new[] { "Customer" }))
      {
        context.Suppliers.FirstOrDefault();
        context.SaveChanges();
      }      
    }

    [TestMethod]
    [ExpectedException(typeof(SecurityException))]
    public void SaveChangesThrowsSecurityExceptionWhenUserIsNotAllowedToInsert()
    {
      using (new TransactionScope())
      using (var context = new NorthwindEntities())
      using (new GenericPrincipalScope("Ed", new[] { "Employee" }))
      {
        context.Suppliers.AddObject(new Supplier { CompanyName = "Test Supplier" });
        context.SaveChanges();
      }
    }

    [TestMethod]
    public void SaveChangesAllowsUserWithValidPermissionToInsert()
    {
      using (new TransactionScope())
      using (var context = new NorthwindEntities())
      using (new GenericPrincipalScope("Mike", new[] { "Manager" }))
      {
        context.Suppliers.AddObject(new Supplier { CompanyName = "Test Supplier" });
        context.SaveChanges();
      }
    }

    [TestMethod]
    [ExpectedException(typeof(SecurityException))]
    public void SaveChangesThrowsSecurityExceptionWhenUserIsNotAllowedToUpdate()
    {
      using (new TransactionScope())
      using (var context = new NorthwindEntities())
      {
        var supplier = new Supplier {CompanyName = "Test Supplier"};

        using (new GenericPrincipalScope("Mike", new[] { "Manager" }))
        {
          context.Suppliers.AddObject(supplier);
          context.SaveChanges();
        }

        using (new GenericPrincipalScope("Ed", new[] { "Employee" }))
        {
          supplier.CompanyName = "Modified by Employee";
          context.SaveChanges();
        }        
      }      
    }

    [TestMethod]
    public void SaveChangesAllowsUserWithValidPermissionToUpdate()
    {
      using (new TransactionScope())
      using (var context = new NorthwindEntities())
      {
        var supplier = new Supplier { CompanyName = "Test Supplier" };

        using (new GenericPrincipalScope("Mike", new[] { "Manager" }))
        {
          context.Suppliers.AddObject(supplier);
          context.SaveChanges();
        }

        using (new GenericPrincipalScope("Michelle", new[] { "Manager" }))
        {
          supplier.CompanyName = "Modified by Manager";
          context.SaveChanges();
        }
      }
    }

    [TestMethod]
    [ExpectedException(typeof(SecurityException))]
    public void SaveChangesThrowsSecurityExceptionWhenUserIsNotAllowedToDelete()
    {
      using (new TransactionScope())
      using (var context = new NorthwindEntities())
      {
        var supplier = new Supplier { CompanyName = "Test Supplier" };

        using (new GenericPrincipalScope("Mike", new[] { "Manager" }))
        {
          context.Suppliers.AddObject(supplier);
          context.SaveChanges();
        }

        using (new GenericPrincipalScope("Ed", new[] { "Employee" }))
        {
          context.Suppliers.DeleteObject(supplier);
          context.SaveChanges();
        }
      }
    }

    [TestMethod]
    public void SaveChangesAllowsUserWithValidPermissionToDelete()
    {
      using (new TransactionScope())
      using (var context = new NorthwindEntities())
      {
        var supplier = new Supplier { CompanyName = "Test Supplier" };

        using (new GenericPrincipalScope("Mike", new[] { "Manager" }))
        {
          context.Suppliers.AddObject(supplier);
          context.SaveChanges();
        }

        using (new GenericPrincipalScope("Michelle", new[] { "Manager" }))
        {
          context.Suppliers.DeleteObject(supplier);
          context.SaveChanges();
        }
      }
    }

    [TestMethod]
    public void CustomersEntitySetFilterOutInaccessibleRecords()
    {
      using (new TransactionScope())
      using (var context = new NorthwindEntities())
      {
        using (new GenericPrincipalScope("Mike", new[] { "Manager" }))
        {
          Assert.IsTrue(context.Customers.Any());
        }

        using (new GenericPrincipalScope("Celene", new[] { "Customer" }))
        {
          Assert.IsFalse(context.Customers.Any());
        }
      }
      
    }
  }
}
