﻿using System;
using System.ComponentModel.DataAnnotations;
using System.Linq;
using System.Transactions;
using Microsoft.VisualStudio.TestTools.UnitTesting;

namespace DataModel
{
  [TestClass]
  public class OrderTest
  {
    private static Order CreateOrder()
    {
      var order = new Order();
      order.OrderStatus = (byte)OrderStatus.Submitted;
      order.OrderDate = DateTime.Today;
      return order;
    }

    [TestMethod]
    [ExpectedValidationException(MemberName = "OrderDate")]
    public void OrderDate_CannotBeInThePast()
    {
      using (new TransactionScope())
      using (var context = new NorthwindEntities())
      using (new GenericPrincipalScope("Mike", new[] { "Manager" }))
      {
        var order = CreateOrder();
        order.OrderDate = DateTime.Today.AddDays(-5);
        context.Orders.AddObject(order);
        context.SaveChanges();
      }
    }

    [TestMethod]
    public void OrderStatus_SetsSubmittedDate()
    {
      using (new TransactionScope())
      using (var context = new NorthwindEntities())
      using (new GenericPrincipalScope("Mike", new[] { "Manager" }))
      {
        var order = new Order();
        order.OrderDate = DateTime.Today;
        order.OrderStatus = (int)OrderStatus.Submitted;
        context.Orders.AddObject(order);
        context.SaveChanges();
        Assert.IsNotNull(order.SubmittedDate);
        Assert.AreEqual(DateTime.Today, order.SubmittedDate.Value.Date);
      }
    }

    [TestMethod]
    [ExpectedValidationException(MemberName = "OrderStatus")]
    public void OrderStatus_RejectsInvalidValues()
    {
      using (new TransactionScope())
      using (var context = new NorthwindEntities())
      using (new GenericPrincipalScope("Mike", new[] { "Manager" }))
      {
        var order = CreateOrder();
        order.OrderStatus = 10;
        context.Orders.AddObject(order);
        context.SaveChanges();
      }
    }

    [TestMethod]
    [ExpectedValidationException(MemberName = "OrderStatus")]
    public void OrderStatus_DoesNotAllowChangingFulfilledToPaid()
    {
      using (new TransactionScope())
      using (var context = new NorthwindEntities())
      using (new GenericPrincipalScope("Mike", new[] { "Manager" }))
      {
        var order = CreateOrder();
        order.OrderStatus = (byte)OrderStatus.Fulfilled;
        context.Orders.AddObject(order);
        context.SaveChanges();
        order.OrderStatus = (byte)OrderStatus.Paid;
        context.SaveChanges();
      }
    }

    [TestMethod]
    public void Validate_ChecksOrderDateAndRequiredDateWhenOrderIsSubmitted()
    {
      using (new TransactionScope())
      using (var context = new NorthwindEntities())
      using (new GenericPrincipalScope("Mike", new[] { "Manager" }))
      {
        var order = new Order();
        order.OrderStatus = (byte)OrderStatus.Submitted;
        order.OrderDate = DateTime.Today.AddDays(-5);
        order.RequiredDate = order.OrderDate;
        context.Orders.AddObject(order);
        try
        {
          context.SaveChanges();
          Assert.Fail();
        }
        catch (AggregateException e)
        {
          Assert.IsTrue(e.InnerExceptions.OfType<ValidationException>().Any(x => x.ValidationResult.MemberNames.Contains("OrderDate")));
          Assert.IsTrue(e.InnerExceptions.OfType<ValidationException>().Any(x => x.ValidationResult.MemberNames.Contains("RequiredDate")));
        }
      }
    }

    [TestMethod]
    [ExpectedValidationException]
    public void Validate_ChecksOutOfStockItems()
    {
      using (new TransactionScope())
      using (var context = new NorthwindEntities())
      using (new GenericPrincipalScope("Mike", new[] { "Manager" }))
      {
        var customer = new Customer();
        customer.CustomerID = "TEST";
        customer.ContactName = "Test Customer";
        customer.CompanyName = "Test Company";
        context.Customers.AddObject(customer);

        var product = new Product();
        product.ProductName = "Test Product";
        product.UnitPrice = 4.95M;
        product.UnitsInStock = 5;
        context.Products.AddObject(product);

        var order = new Order();
        order.Customer = customer;
        order.OrderDate = DateTime.Today;
        order.OrderStatus = (byte)OrderStatus.Draft;
        context.Orders.AddObject(order);

        var item = new Order_Detail();
        item.Product = product;
        item.Quantity = 1;
        order.Order_Details.Add(item);

        context.SaveChanges();

        product.UnitsInStock = 0;
        context.SaveChanges();

        order.OrderStatus = (byte)OrderStatus.Submitted;
        context.SaveChanges();
      }
    }

    [TestMethod]
    public void SubmittedOrderCannotBeDeleted()
    {
      using (new TransactionScope())
      using (var context = new NorthwindEntities())
      using (new GenericPrincipalScope("Mike", new[] { "Manager" }))
      {
        var order = new Order();
        order.OrderDate = DateTime.Today;
        order.OrderStatus = (int)OrderStatus.Submitted;
        context.Orders.AddObject(order);
        context.SaveChanges();

        context.Orders.DeleteObject(order);
        try
        {
          context.SaveChanges();
          Assert.Fail();
        }
        catch (ValidationException)
        {
          // this is what we expect
        }
      }
    }

    [TestMethod]
    public void Fulfill_UpdatesProductUnitsInStock()
    {
      var product = new Product();
      product.UnitsInStock = 5;

      var orderItem = new Order_Detail();
      orderItem.Product = product;
      orderItem.Quantity = 2;

      var order = new Order();
      order.OrderStatus = (byte)OrderStatus.Paid;
      order.Order_Details.Add(orderItem);

      order.Fulfill();
      Assert.AreEqual((byte)OrderStatus.Fulfilled, order.OrderStatus);
      Assert.AreEqual(3, product.UnitsInStock.Value);
    }
  }
}
