﻿namespace DataModel
{
  using System;
  using System.Collections.Generic;
  using System.ComponentModel.DataAnnotations;
  using System.Data;
  using System.Data.Objects;
  using System.Diagnostics;
  using System.Linq;
  using System.Security.Principal;
  using System.Threading;
  using Unleashed.DataAnnotations;
  using Unleashed.EntityFramework;

  [MetadataType(typeof(Metadata))]
  partial class Order : IValidatableObject, ISavableObject
  {
    public void BeforeSave(ObjectStateEntry stateEntry)
    {
      switch (stateEntry.State)
      {
        case EntityState.Added:
        case EntityState.Modified:
          byte? oldOrderStatus = (byte)DataModel.OrderStatus.Draft;
          if (stateEntry.State == EntityState.Modified)
            oldOrderStatus = stateEntry.OriginalValues.Get(() => this.OrderStatus);

          if (this.OrderStatus == (byte)DataModel.OrderStatus.Submitted &&
              oldOrderStatus == (byte)DataModel.OrderStatus.Draft)
          {
            this.SubmittedDate = DateTime.Now;
          }

          return;

        case EntityState.Deleted:
          if (this.OrderStatus != (byte)DataModel.OrderStatus.Draft)
            throw new ValidationException("Only draft orders can be deleted");
          return;
      }
    }

    public void Fulfill()
    {
      if (this.OrderStatus != (byte)DataModel.OrderStatus.Paid)
      {
        throw new ValidationException(
          new ValidationResult("Only paid orders can be fulfilled", new[] { "OrderStatus" }),
          null, this.OrderStatus);
      }

      this.OrderStatus = (byte)DataModel.OrderStatus.Fulfilled;
      this.ShippedDate = DateTime.Now;
      foreach (Order_Detail orderItem in this.Order_Details)
      {
        orderItem.Product.UnitsInStock -= orderItem.Quantity;
      }
    }

    public IEnumerable<ValidationResult> Validate(ValidationContext context)
    {
      switch ((OrderStatus?)this.OrderStatus)
      {
        case null:
        case DataModel.OrderStatus.Draft:
          return this.ValidateDraftOrder(context);

        case DataModel.OrderStatus.Submitted:
          return this.ValidateSubmittedOrder(context);

        case DataModel.OrderStatus.Paid:
          return this.ValidatePaidOrder(context);

        case DataModel.OrderStatus.Fulfilled:
          return this.ValidateFulfilledOrder(context);

        default:
          Debug.Fail("Unexpected OrderStatus value");
          return new ValidationResult[0];
      }
    }

    private IEnumerable<ValidationResult> ValidateDraftOrder(ValidationContext context)
    {
      // Validation logic specific to draft orders
      if (this.Customer == null && string.IsNullOrEmpty(this.CustomerID))
        yield return new ValidationResult("The Customer is required", new[] { "Customer" });
    }

    private IEnumerable<ValidationResult> ValidateSubmittedOrder(ValidationContext context)
    {
      if (this.OrderDate == null)
        yield return new ValidationResult("The Order Date is required", new[] { "OrderDate" });

      if (this.OrderDate.Value < DateTime.Today)
        yield return new ValidationResult("The Order Date cannot be in the past", new [] { "OrderDate" });

      if (this.RequiredDate != null && this.RequiredDate <= this.OrderDate.Value.AddDays(3))
        yield return new ValidationResult("The Required Date must be greater than Order Date + 3 days", new[] { "RequiredDate" });

      var entities = (NorthwindEntities)context.Items[typeof(ObjectContext)];
      string productName = entities.Order_Details
        .Where(item => item.OrderID == this.OrderID
                    && item.Product.UnitsInStock < item.Quantity)
        .Select(item => item.Product.ProductName).FirstOrDefault();
      if (productName != null)
        yield return new ValidationResult(productName + " is out of stock");
    }

    private IEnumerable<ValidationResult> ValidatePaidOrder(ValidationContext context)
    {
      // Validation logic specific to paid orders
      return new ValidationResult[0];
    }

    private IEnumerable<ValidationResult> ValidateFulfilledOrder(ValidationContext context)
    {
      // Validation logic specific to fulfilled orders
      return new ValidationResult[0];
    }

    public static ValidationResult ValidateOrderStatus(byte? newStatus, ValidationContext context)
    {
      if (context != null)
      {
        var stateEntry = (ObjectStateEntry)context.Items[typeof(ObjectStateEntry)];
        if (stateEntry.State == EntityState.Modified)
        {
          var order = (Order)stateEntry.Entity;
          byte? oldStatus = stateEntry.OriginalValues.Get(() => order.OrderStatus);
          if (oldStatus == (byte?)DataModel.OrderStatus.Fulfilled && newStatus != (byte?)DataModel.OrderStatus.Fulfilled)
          {
            return new ValidationResult("Status of fulfilled orders should not be modified", new string[] { "OrderStatus" });
          }
        }
      }
      
      return ValidationResult.Success;
    }

    public static IQueryable GetQuery(IQueryable source)
    {
      IPrincipal user = Thread.CurrentPrincipal;
      if (user.Identity.IsAuthenticated == false)
        source = source.Cast<Order>().Where(order => false);
      else if (user.IsInRole("Customer"))
        source = source.Cast<Order>().Where(order => order.Customer.EmailAddress == user.Identity.Name);
      return source;
    }

    [Authorization(
      Allow.Roles, "Manager",
      Allow.Roles, "Employee", Actions.Read | Actions.Update | Actions.Insert,
      Allow.Roles, "Customer", Actions.Read,
      Deny.Users, "*")]
    [CustomQuery(typeof(Order), "GetQuery")]
    private abstract class Metadata
    {
      [Required]
      [FilterUIHint("Date")]
      [DisplayFormat(DataFormatString = "{0:D}")]
      [Display(Name = "Order Date")]
      public object OrderDate { get; set; }

      [DataType(DataType.Date)]
      [Display(Name = "Required Date")]
      //[Editable(false, AllowInitialValue = true)]
      public object RequiredDate { get; set; }

      [Display(Name = "Shipped Date")]
      public object ShippedDate { get; set; }

      [Display(Name = "Ship Name")]
      public object ShipName { get; set; }

      [Display(Name = "Ship Address")]
      public object ShipAddress { get; set; }

      [Display(Name = "Ship City")]
      public object ShipCity { get; set; }

      [Display(Name = "Ship Region")]
      public object ShipRegion { get; set; }

      [Display(Name = "Ship Postal Code")]
      public object ShipPostalCode { get; set; }

      [Display(Name = "Ship Country")]
      public object ShipCountry { get; set; }

      [Display(Name = "Order Status")]
      [EnumDataType(typeof(OrderStatus))]
      [CustomValidation(typeof(Order), "ValidateOrderStatus")]
      public object OrderStatus { get; set; }

      //[Editable(false)]
      public object SubmittedDate { get; set; }
    }
  }
}
