﻿namespace Unleashed.DataAnnotations
{
  using System;
  using System.Collections.Generic;
  using System.Security.Principal;
  using System.Collections.ObjectModel;

  [AttributeUsage(AttributeTargets.Class)]
  public class AuthorizationAttribute : Attribute
  {
    private readonly Lazy<ReadOnlyCollection<AuthorizationRule>> rules;

    public AuthorizationAttribute(params object[] rules)
    {
      this.rules = new Lazy<ReadOnlyCollection<AuthorizationRule>>(() => ParseRules(rules));
    }

    public ReadOnlyCollection<AuthorizationRule> Rules
    {
      get { return this.rules.Value; }
    }

    public bool IsAuthorized(IPrincipal principal, Actions actions = Actions.All)
    {
      Actions allowedActions = Actions.None;

      foreach (AuthorizationRule rule in this.Rules)
      {
        if (!rule.IsMatch(principal)) continue;

        if (rule.Authorize)
        {
          allowedActions |= rule.Actions;
          if ((allowedActions & actions) == actions)
            return true;
        }
        else
        {
          if ((rule.Actions & actions) != Actions.None)
            return false;          
        }
      }

      return true; // mimic the default <allow users="*"/> rule
    }

    private static ReadOnlyCollection<AuthorizationRule> ParseRules(IEnumerable<object> tokens)
    {
      var rules = new List<AuthorizationRule>();

      object rule = null;
      List<string> names = null;
      Actions? actions = null;
      foreach (object token in tokens)
      {
        if (token is Allow || token is Deny)
        {
          StartNewRule(rules, ref rule, ref names, ref actions);
          rule = token;
        }
        else if (token is string)
        {
          ThrowIfRuleNotSpecified(rule);
          names.Add((string)token);
        }
        else if (token is Actions)
        {
          ThrowIfRuleNotSpecified(rule);
          actions = actions ?? Actions.None;
          actions |= (Actions)token;
        }
      }

      SaveLastRule(rules, rule, names, actions);
      return new ReadOnlyCollection<AuthorizationRule>(rules);
    }

    private static void StartNewRule(List<AuthorizationRule> rules, ref object rule, ref List<string> names, ref Actions? actions)
    {
      if (rule != null)
        SaveLastRule(rules, rule, names, actions);

      rule = null;
      names = new List<string>();
      actions = null;
    }

    private static void SaveLastRule(List<AuthorizationRule> rules, object rule, List<string> names, Actions? actions)
    {
      ThrowIfRuleNotSpecified(rule);

      bool authorize = rule is Allow;

      bool isUserBased = rule.Equals(Allow.Users) || rule.Equals(Deny.Users);
      List<string> users = isUserBased ? names : new List<string>();
      List<string> roles = isUserBased ? new List<string>() : names;

      actions = actions ?? Actions.All;

      rules.Add(new AuthorizationRule(authorize, actions.Value, roles, users));      
    }

    private static void ThrowIfRuleNotSpecified(object rule)
    {
      if (rule == null)
        throw new ArgumentException("Authorization rule must begin with Allow.Users, Allow.Roles, Deny.Users or Deny.Roles");
    }
  }
}
