﻿namespace Unleashed.DynamicData.Tests
{
  using System;
  using System.Collections.Generic;
  using System.Linq;
  using System.Security.Principal;
  using System.Web.DynamicData;
  using System.Web.DynamicData.ModelProviders;
  using System.Web.UI.WebControls;
  using DataAnnotations;
  using Microsoft.VisualStudio.TestTools.UnitTesting;

  [TestClass]
  public class UnleashedMetaTableTest
  {
    private class EntityWithoutAuthorizationAttributes {}

    [Authorization(Deny.Users, "?")]
    private class EntityWithAnonymousAccessDenied { }

    [CustomQueryAttribute(typeof(FakeEntityWithGetQueryMethod), "GetQuery")]
    private class FakeEntityWithGetQueryMethod
    {
      public static Func<IQueryable, IQueryable> GetQueryMethod;
      public static IQueryable GetQuery(IQueryable source)
      {
        return GetQueryMethod(source);
      }
    }

    [TestMethod]
    public void CanReadReturnsTrueByDefault()
    {
      UnleashedMetaTable metaTable = this.CreateSimpleMetaTable(typeof(EntityWithoutAuthorizationAttributes));
      Assert.IsTrue(metaTable.CanRead(new GenericPrincipal(new GenericIdentity(string.Empty), null)));
    }

    [TestMethod]
    public void CanReadReturnsFalseWhenUserIsDeniedAccess()
    {
      UnleashedMetaTable metaTable = this.CreateSimpleMetaTable(typeof(EntityWithAnonymousAccessDenied));
      Assert.IsFalse(metaTable.CanRead(new GenericPrincipal(new GenericIdentity(string.Empty), null)));
    }

    [TestMethod]
    public void CanInsertReturnsTrueByDefault()
    {
      UnleashedMetaTable metaTable = this.CreateSimpleMetaTable(typeof(EntityWithoutAuthorizationAttributes));
      Assert.IsTrue(metaTable.CanInsert(new GenericPrincipal(new GenericIdentity(string.Empty), null)));
    }

    [TestMethod]
    public void CanInsertReturnsFalseWhenUserIsDeniedAccess()
    {
      UnleashedMetaTable metaTable = this.CreateSimpleMetaTable(typeof(EntityWithAnonymousAccessDenied));
      Assert.IsFalse(metaTable.CanInsert(new GenericPrincipal(new GenericIdentity(string.Empty), null)));
    }

    [TestMethod]
    public void CanUpdateReturnsTrueByDefault()
    {
      UnleashedMetaTable metaTable = this.CreateSimpleMetaTable(typeof(EntityWithoutAuthorizationAttributes));
      Assert.IsTrue(metaTable.CanUpdate(new GenericPrincipal(new GenericIdentity(string.Empty), null)));
    }

    [TestMethod]
    public void CanUpdateReturnsFalseWhenUserIsDeniedAccess()
    {
      UnleashedMetaTable metaTable = this.CreateSimpleMetaTable(typeof(EntityWithAnonymousAccessDenied));
      Assert.IsFalse(metaTable.CanUpdate(new GenericPrincipal(new GenericIdentity(string.Empty), null)));
    }

    [TestMethod]
    public void CanDeleteReturnsTrueByDefault()
    {
      UnleashedMetaTable metaTable = this.CreateSimpleMetaTable(typeof(EntityWithoutAuthorizationAttributes));
      Assert.IsTrue(metaTable.CanDelete(new GenericPrincipal(new GenericIdentity(string.Empty), null)));
    }

    [TestMethod]
    public void CanDeleteReturnsFalseWhenUserIsDeniedAccess()
    {
      UnleashedMetaTable metaTable = this.CreateSimpleMetaTable(typeof(EntityWithAnonymousAccessDenied));
      Assert.IsFalse(metaTable.CanDelete(new GenericPrincipal(new GenericIdentity(string.Empty), null)));
    }

    [TestMethod]
    public void GetQueryableUsesCustomQueryAttribute()
    {
      // Prepare Fake Entity and its GetQuery method      
      var entityQuery = new List<string>().AsQueryable();
      FakeEntityWithGetQueryMethod.GetQueryMethod = query => entityQuery;

      // Create MetaTable object
      var simpleTableProvider = MetaTable.CreateTable(typeof(FakeEntityWithGetQueryMethod)).Provider;
      var metaTable = new UnleashedMetaTable(null, simpleTableProvider);
      
      // Call the GetQueryable method
      var sourceQuery = new List<string>().AsQueryable();
      IQueryable resultQuery = metaTable.GetQueryable(sourceQuery);

      // Verify that UnleashedMetaTable.GetQueryable return the query supplied by the Fake Entity's GetQuery method
      Assert.AreSame(entityQuery, resultQuery);
    }

    [TestMethod]
    public void GetQueryUsesCustomQueryAttribute()
    {
      // Get the built-in simple metadata provider for the test entity
      var simpleTableProvider = MetaTable.CreateTable(typeof(FakeEntityWithGetQueryMethod)).Provider;

      // Fake the CreateContext method of the DataModelProvider to prevent the NotSupportedException
      var fakeModelProvider = new FakeDataModelProvider(simpleTableProvider.DataModel)
        {
          CreateContextMethod = () => null
        };

      // Fake the GetQuery method of the TableProvider to have it return a test query object
      var providerQuery = new List<string>().AsQueryable();
      var fakeTableProvider = new FakeTableProvider(fakeModelProvider, simpleTableProvider)
        {
          GetQueryMethod = contextObject => providerQuery
        };

      // Verify that UnleashedMetaTable.GetQuery passes the provider query object to the entity method
      var entityQuery = new List<string>().AsQueryable();
      FakeEntityWithGetQueryMethod.GetQueryMethod = sourceQuery =>
        {
          Assert.AreSame(providerQuery, sourceQuery);
          return entityQuery;
        };

      // Verify that UnleashedMetaTable.GetQuery returns the query object returned by the entity method
      var unleashedMetaTable = new UnleashedMetaTable(null, fakeTableProvider);
      IQueryable tableQuery = unleashedMetaTable.GetQuery(null);
      Assert.AreSame(entityQuery, tableQuery);
    }

    [TestMethod]
    public void GetScaffoldColumnsDoesNotReturnChildrenColumnsUserCannotRead()
    {
      var modelProvider = new FakeDataModelProvider(null);

      var childTableProvider = new FakeTableProvider(modelProvider, null) {NameGetter = () => "Child"};

      var parentTableProvider = new FakeTableProvider(modelProvider, null) {NameGetter = () => "Parent"};
      var nameColumnProvider = new FakeColumnProvider(parentTableProvider) {NameGetter = () => "Name"};
      var childrenColumnProvider = new FakeColumnProvider(parentTableProvider) {NameGetter = () => "Children"};
      childrenColumnProvider.AssociationGetter = () => new FakeAssociationProvider
        {
          DirectionGetter = () => AssociationDirection.OneToMany,
          ToTableGetter = () => childTableProvider,
          FromColumnGetter = () => childrenColumnProvider
        };
      parentTableProvider.ColumnsGetter = () => new[] {nameColumnProvider, childrenColumnProvider};

      modelProvider.TablesGetter = () => new[] {childTableProvider, parentTableProvider};

      var model = new UnleashedMetaModel(false);
      model.RegisterContext(modelProvider);

      var parentTable = (UnleashedMetaTable)model.GetTable("Parent");
      var childrenColumn = (MetaChildrenColumn)parentTable.GetColumn("Children");
      var childTable = (UnleashedMetaTable)model.GetTable("Child");
      Assert.AreSame(childTable, childrenColumn.ChildTable);

      HttpContext.Current = new FakeHttpContext();
      HttpContext.Current.User = new GenericPrincipal(new GenericIdentity(string.Empty), null);

      // By default, Children column is scaffolded
      List<MetaColumn> columns = parentTable.GetScaffoldColumns(DataBoundControlMode.Edit, ContainerType.List).ToList();
      CollectionAssert.Contains(columns, childrenColumn);

      // When user cannot read the child table then Children column is not scaffolded
      childTableProvider.CanReadMethod = principal => false;
      columns = parentTable.GetScaffoldColumns(DataBoundControlMode.Edit, ContainerType.List).ToList();
      CollectionAssert.DoesNotContain(columns, childrenColumn);
    }

    private UnleashedMetaTable CreateSimpleMetaTable(Type entityType)
    {
      var metaTable = MetaTable.CreateTable(entityType);
      var model = new UnleashedMetaModel(false);
      model.RegisterContext(metaTable.Provider.DataModel);
      return model.Tables.Cast<UnleashedMetaTable>().First();
    }
  }
}