namespace Unleashed.DynamicData
{
  using System;
  using System.Collections.Specialized;
  using System.ComponentModel;
  using System.Web.DynamicData;
  using System.Web.UI;
  using System.Web.UI.WebControls;

  public class UnleashedFieldTemplate : FieldTemplateUserControl
  {
    private static readonly object FieldValueChangedEvent = new object();

    public virtual bool AutoPostBack { get; set; }

    public override object FieldValue
    {
      get { return base.FieldValue; }
      set
      {
        if (base.FieldValue != value)
        {
          base.FieldValue = value;
          this.OnFieldValueChanged(EventArgs.Empty);
        }
      }
    }

    protected override object GetColumnValue(MetaColumn column)
    {
      try
      {
        return base.GetColumnValue(column);        
      }
      catch (InvalidOperationException)
      {
        return this.ExtractFieldValue();
      }
    }

    public event EventHandler FieldValueChanged
    {
      add { this.Events.AddHandler(FieldValueChangedEvent, value); }
      remove { this.Events.RemoveHandler(FieldValueChangedEvent, value); }
    }

    protected new UnleashedFieldTemplate FindOtherFieldTemplate(string columnName)
    {
      var template = (UnleashedFieldTemplate)base.FindOtherFieldTemplate(columnName);

      // If other field template is inside of a naming container
      if (template == null && !string.IsNullOrEmpty(columnName))
      {
        string controlId = "__" + columnName;
        Control parent = this.FindParentOfAllFieldTemplates();
        template = (UnleashedFieldTemplate)FindControlRecursively(parent, controlId);
      }

      return template;
    }

    private Control FindParentOfAllFieldTemplates()
    {
      Control parent = this.Parent;
      while (parent != null)
      {
        if (parent is GridViewRow || parent is FormViewRow || parent is DetailsViewRow) 
          break;
        parent = parent.Parent;
      }

      return parent;
    }

    private static Control FindControlRecursively(Control parent, string controlId)
    {
      Control control = parent.FindControl(controlId);
      if (control != null) 
        return control;

      foreach (Control child in parent.Controls)
      {
        control = FindControlRecursively(child, controlId);
        if (control != null && string.Equals(control.ID, controlId, StringComparison.OrdinalIgnoreCase)) // workaround for the quirk in CheckBoxList and RadioButtonList controls
          return control;
      }        

      return null;
    }

    protected virtual void OnFieldValueChanged(EventArgs e)
    {
      var eventHandler = (EventHandler)this.Events[FieldValueChangedEvent];
      if (eventHandler != null)
        eventHandler(this, e);
    }

    protected void SetUpValidator(object sender, EventArgs args)
    {
      var validator = (BaseValidator)sender;
      if (validator is DynamicValidator)
      {
        validator.Enabled = true; // By default only DynamicValidators are enabled. The base class will do the rest.
        validator.PreRender += this.DynamicValidator_PreRender;
      }
      validator.SetFocusOnError = true; // This cannot be set via skin, useful for large pages
      this.SetUpValidator(validator);
      validator.Visible = validator.Enabled;
    }

    private void DynamicValidator_PreRender(object sender, EventArgs e)
    {
      var validator = (DynamicValidator)sender;
      validator.ToolTip = validator.ErrorMessage;
    }

    private object ExtractFieldValue()
    {
      var dictionary = new OrderedDictionary();
      this.ExtractValues(dictionary);
      if (this.Column is MetaForeignKeyColumn)
      {
        var values = new object[dictionary.Count];
        dictionary.Values.CopyTo(values, 0);
        return string.Join(",", values);
      }

      TypeConverter converter = TypeDescriptor.GetConverter(this.Column.ColumnType);
      return converter.ConvertFrom(dictionary[this.Column.Name]);
    }
  }
}