namespace Unleashed.DynamicData
{
  using System.Collections.Generic;
  using System.ComponentModel.DataAnnotations;
  using System.Linq;
  using System.Security.Principal;
  using System.Web.DynamicData;
  using System.Web.DynamicData.ModelProviders;
  using System.Web.UI.WebControls;
  using DataAnnotations;

  public class UnleashedMetaTable : MetaTable
  {
    private readonly AuthorizationAttribute authorizationAttribute;
    private readonly CustomQueryAttribute customQueryAttribute;

    public UnleashedMetaTable(UnleashedMetaModel model, TableProvider provider)
      : base(model, provider)
    {
      this.authorizationAttribute = this.Attributes.OfType<AuthorizationAttribute>().FirstOrDefault();
      this.customQueryAttribute = this.Attributes.OfType<CustomQueryAttribute>().FirstOrDefault();
    }

    public override bool CanDelete(IPrincipal principal)
    {
      return base.CanDelete(principal) && this.IsAuthorized(principal, Actions.Delete);
    }

    public override bool CanInsert(IPrincipal principal)
    {
      return base.CanInsert(principal) && this.IsAuthorized(principal, Actions.Insert);
    }

    public override bool CanRead(IPrincipal principal)
    {
      return base.CanRead(principal) && this.IsAuthorized(principal, Actions.Read);
    }

    public override bool CanUpdate(IPrincipal principal)
    {
      return base.CanUpdate(principal) && this.IsAuthorized(principal, Actions.Update);
    }

    public override IEnumerable<MetaColumn> GetFilteredColumns()
    {
      return this.GetScaffoldColumns(DataBoundControlMode.ReadOnly, ContainerType.List)
        .Where(column => this.IsFilterable(column));
    }

    public IQueryable GetQueryable(IQueryable query)
    {
      if (this.customQueryAttribute != null)
        query = this.customQueryAttribute.GetQueryable(query);
      return query;
    }

    public override IQueryable GetQuery(object context)
    {
      IQueryable baseQuery = base.GetQuery(context);
      IQueryable customQuery = this.GetQueryable(baseQuery);
      return customQuery;
    }

    public override IEnumerable<MetaColumn> GetScaffoldColumns(DataBoundControlMode mode, ContainerType containerType)
    {
      return base.GetScaffoldColumns(mode, containerType).Where(column => IsScaffoldable(column));
    }

    private bool IsAuthorized(IPrincipal principal, Actions actions)
    {
      if (this.authorizationAttribute == null)
        return true;
      return this.authorizationAttribute.IsAuthorized(principal, actions);
    }

    private bool IsFilterable(MetaColumn column)
    {
      if (!string.IsNullOrEmpty(column.FilterUIHint)) return true;

      DisplayAttribute display = column.Attributes.OfType<DisplayAttribute>().FirstOrDefault();
      if (display != null && display.GetAutoGenerateFilter().HasValue)
        return display.GetAutoGenerateFilter().Value;

      if (column.IsCustomProperty) return false;
      if (column is MetaChildrenColumn) return false;
      return true;
    }

    private bool IsScaffoldable(MetaColumn column)
    {
      var childrenColumn = column as MetaChildrenColumn;
      if (childrenColumn != null)
        return childrenColumn.ChildTable.CanRead(HttpContext.Current.User);
      return true;      
    }
  }
}