using System;
using System.Collections.Concurrent;
using System.Collections.Generic;
using System.ComponentModel;
using System.ComponentModel.DataAnnotations;
using System.Data;
using System.Data.EntityClient;
using System.Data.Objects;
using System.Linq;
using System.Security;
using System.Security.Principal;
using System.Threading;
using Unleashed.DataAnnotations;

namespace Unleashed.EntityFramework
{
  public class UnleashedObjectContext : ObjectContext
  {
    private const EntityState ChangedEntityStates = EntityState.Added | EntityState.Modified | EntityState.Deleted;

    private static readonly ConcurrentDictionary<Type, AuthorizationAttribute> AuthorizationAttributes = new ConcurrentDictionary<Type, AuthorizationAttribute>();

    protected UnleashedObjectContext(string connectionString, string defaultContainerName) : base(connectionString, defaultContainerName)
    {
    }

    protected UnleashedObjectContext(EntityConnection connection, string defaultContainerName) : base(connection, defaultContainerName)
    {
    }

    protected static void Authorize(Type entityType, Actions action)
    {
      IPrincipal user = Thread.CurrentPrincipal;
      if (!IsAuthorized(user, entityType, action))
      {
        throw new SecurityException(string.Format(
          "User '{0}' is not authorized to perform action '{1}' on entities of type '{2}'",
          user.Identity.Name, action, entityType.Name));
      }
    }

    private static Actions EntityStateToAction(EntityState state)
    {
      switch (state)
      {
        case EntityState.Added:
          return Actions.Insert;
        case EntityState.Modified:
          return Actions.Update;
        case EntityState.Deleted:
          return Actions.Delete;
        default:
          throw new ArgumentException("state");
      }
    }

    private static bool IsAuthorized(IPrincipal user, Type entityType, Actions action)
    {
      AuthorizationAttribute authorizationAttribute = AuthorizationAttributes.GetOrAdd(entityType,
        type =>
          {
            Activator.CreateInstance(type); // force execution of the static constructor to initialize metadata
            return TypeDescriptor.GetAttributes(type).OfType<AuthorizationAttribute>().FirstOrDefault();
          });

      return authorizationAttribute == null || authorizationAttribute.IsAuthorized(user, action);
    }

    public override int SaveChanges(SaveOptions options)
    {
      IEnumerable<ObjectStateEntry> stateEntries = this.ObjectStateManager
        .GetObjectStateEntries(ChangedEntityStates)
        .Where(stateEntry => !stateEntry.IsRelationship);

      foreach (ObjectStateEntry stateEntry in stateEntries)
        this.BeforeSaveChanges(stateEntry);

      return base.SaveChanges(options);
    }

    private void BeforeSaveChanges(ObjectStateEntry stateEntry)
    {
      Authorize(stateEntry.Entity.GetType(), EntityStateToAction(stateEntry.State));

      var savable = stateEntry.Entity as ISavableObject;
      if (savable != null)
        savable.BeforeSave(stateEntry);

      if (stateEntry.State != EntityState.Deleted)
        this.Validate(stateEntry);
    }

    protected void Validate(object entity)
    {
      this.Validate(this.ObjectStateManager.GetObjectStateEntry(entity));
    }

    private void Validate(ObjectStateEntry stateEntry)
    {
      object entity = stateEntry.Entity;

      var items = new Dictionary<object, object>();
      items[typeof(ObjectStateEntry)] = stateEntry;
      items[typeof(ObjectContext)] = this;

      var context = new ValidationContext(entity, null, items);

      var results = new List<ValidationResult>();
      if (!Validator.TryValidateObject(entity, context, results, true))
        ThrowValidationException(entity, results);
    }

    protected static void ThrowValidationException(object entity, IEnumerable<ValidationResult> results)
    {
      var exceptions = new List<ValidationException>();
      foreach (ValidationResult result in results)
        exceptions.Add(new ValidationException(result, null, entity));

      if (exceptions.Count == 1)
        throw exceptions[0];

      throw new AggregateException(exceptions);
    }
  }
}